from sqlalchemy.sql.expression import text, TextClause
from records_mover.utils import quiet_remove
from records_mover.records.delimited import cant_handle_hint
from records_mover.records import DelimitedRecordsFormat
from records_mover.records.delimited import (
    HintEncoding, HintRecordTerminator,
    HintFieldDelimiter, HintQuoteChar,
    HintQuoting, HintEscape,
    HintHeaderRow, HintCompression,
    HintDoublequote,
)
from ...records.delimited.validated_records_hints import ValidatedRecordsHints
from typing import (Optional, Set, Dict, NamedTuple,
                    NoReturn, Tuple, TYPE_CHECKING)
if TYPE_CHECKING:
    from typing_extensions import Literal
    # http://dev.mysql.com/doc/refman/8.0/en/charset-mysql.html
    MySqlCharacterSet = Literal['big5', 'binary', 'latin1', 'ucs2',
                                'utf8', 'utf8mb4', 'utf16', 'utf16le',
                                'utf32']
else:
    MySqlCharacterSet = str


# The NoReturn annotations are the mypy way of validating we're
# covering all officially approved cases of an the hint at
# typechecking time.
#
# https://github.com/python/mypy/issues/6366#issuecomment-560369716
def _assert_never(the_bad_hint: NoReturn) -> NoReturn:
    assert False


MYSQL_CHARACTER_SETS_FOR_LOAD: Dict[HintEncoding, MySqlCharacterSet] = {
    "UTF8": 'utf8',
    "UTF16": 'utf16',
    "UTF16LE": 'utf16le',
    # https://dev.mysql.com/doc/refman/8.0/en/charset-unicode.html
    #
    # The MySQL implementation of UCS-2, UTF-16, and UTF-32 stores
    # characters in big-endian byte order and does not use a byte
    # order mark (BOM) at the beginning of values. Other database
    # systems might use little-endian byte order or a BOM. In such
    # cases, conversion of values will need to be performed when
    # transferring data between those systems and MySQL. The
    # implementation of UTF-16LE is little-endian.
    "UTF16BE": 'utf16',
    # MySQL uses no BOM for UTF-8 values.
    "LATIN1": 'latin1',
    #  MySQL's latin1 is the same as the Windows cp1252 character set.
    #
    # https://dev.mysql.com/doc/refman/8.0/en/charset-we-sets.html
    "CP1252": 'latin1',
}


# Mark as total=False so we can create this incrementally
class MySqlLoadOptions(NamedTuple):
    character_set: MySqlCharacterSet
    fields_terminated_by: str  # default '\t'
    fields_enclosed_by: Optional[str]  # default None
    fields_optionally_enclosed_by: Optional[str]  # default None
    fields_escaped_by: Optional[str]  # default '\'
    lines_starting_by: str  # default ''
    lines_terminated_by: str  # default '\n'
    ignore_n_lines: int

    def generate_load_data_sql(self,
                               filename: str,
                               schema_name: str,
                               table_name: str) -> TextClause:
        sql = f"""\
LOAD DATA
LOCAL INFILE :filename
INTO TABLE {schema_name}.{table_name}
CHARACTER SET :character_set
FIELDS
    TERMINATED BY :fields_terminated_by
"""
        if self.fields_enclosed_by is not None and self.fields_optionally_enclosed_by is None:
            sql += "    ENCLOSED BY :fields_enclosed_by\n"
        elif self.fields_enclosed_by is None and self.fields_optionally_enclosed_by is not None:
            sql += "    OPTIONALLY ENCLOSED BY :fields_optionally_enclosed_by\n"
        elif self.fields_enclosed_by is not None and self.fields_optionally_enclosed_by is not None:
            raise SyntaxError('fields_enclosed_by and optionally_fields_enclosed_by '
                              'cannot both be set')
        if self.fields_escaped_by is not None:
            sql += "    ESCAPED BY :fields_escaped_by\n"
        sql += """\
LINES
    STARTING BY :lines_starting_by
    TERMINATED BY :lines_terminated_by
IGNORE :ignore_n_lines LINES
"""
        clause = text(sql)
        clause = clause.\
            bindparams(filename=filename,
                       character_set=self.character_set,
                       fields_terminated_by=self.fields_terminated_by,
                       lines_starting_by=self.lines_starting_by,
                       lines_terminated_by=self.lines_terminated_by,
                       ignore_n_lines=self.ignore_n_lines)
        if self.fields_enclosed_by is not None:
            clause = clause.bindparams(fields_enclosed_by=self.
                                       fields_enclosed_by)
        if self.fields_optionally_enclosed_by is not None:
            clause = clause.bindparams(fields_optionally_enclosed_by=self.
                                       fields_optionally_enclosed_by)
        if self.fields_escaped_by is not None:
            clause = clause.bindparams(fields_escaped_by=self.
                                       fields_escaped_by)
        return clause


class MysqlLoadParameters:
    """
    A class representing MySQL load parameters.

    Attributes:
        hints (ValidatedRecordsHints): Validated records hints.
        hint_quoting (HintQuoting): Quoting hint.
        fail_if_cant_handle_hint (bool): Fail if unable to handle hint.
        records_format (DelimitedRecordsFormat): Records format.
        unhandled_hints (Set[str]): Set of unhandled hints.
    """
    def __init__(self,
                 fail_if_cant_handle_hint: bool,
                 records_format: DelimitedRecordsFormat,
                 unhandled_hints: Set[str]):
        """Initialize the MySQL load parameters."""
        self.hints: ValidatedRecordsHints = records_format.validate(
            fail_if_cant_handle_hint=fail_if_cant_handle_hint)
        self.hint_quoting: HintQuoting = self.hints.quoting
        self.fail_if_cant_handle_hint: bool = fail_if_cant_handle_hint
        self.records_format: DelimitedRecordsFormat = records_format
        self.unhandled_hints: Set[str] = unhandled_hints


def get_character_set(mysql_load_parameters: MysqlLoadParameters) -> MySqlCharacterSet:
    """
    Determine the character set for the given MySQL load parameters.
    The server uses the character set indicated by the
    character_set_database system variable to interpret the
    information in the file. SET NAMES and the setting of
    character_set_client do not affect interpretation of input. If
    the contents of the input file use a character set that differs
    from the default, it is usually preferable to specify the
    character set of the file by using the CHARACTER SET clause. A
    character set of binary specifies “no conversion.”

    Args:
        mysql_load_parameters (MysqlLoadParameters): MySQL load parameters.

    Returns:
        MySqlCharacterSet: MySQL character set.
    """

    hint_encoding: HintEncoding = mysql_load_parameters.hints.encoding
    character_set = MYSQL_CHARACTER_SETS_FOR_LOAD.get(hint_encoding)
    if character_set is not None:
        mysql_character_set = character_set
    else:
        cant_handle_hint(mysql_load_parameters.fail_if_cant_handle_hint,
                         'encoding',
                         mysql_load_parameters.records_format.hints)
        mysql_character_set = 'utf8'
    return mysql_character_set


def get_field_terminator(mysql_load_parameters: MysqlLoadParameters
                         ) -> str:
    """
    Get the field terminator for the given MySQL load parameters.

    Args:
        mysql_load_parameters (MysqlLoadParameters): MySQL load parameters.

    Returns:
        str: Field terminator.
    """

    field_terminator: HintFieldDelimiter = mysql_load_parameters.hints.field_delimiter
    mysql_fields_terminator = field_terminator
    return mysql_fields_terminator


def get_enclosures(mysql_load_parameters: MysqlLoadParameters
                   ) -> Tuple[Optional[str], Optional[str]]:
    """
    Get the enclosures for the given MySQL load parameters.

    https://dev.mysql.com/doc/refman/8.0/en/load-data.html

    LOAD DATA can be used to read files obtained from external
    sources. For example, many programs can export data in
    comma-separated values (CSV) format, such that lines have fields
    separated by commas and enclosed within double quotation marks,
    with an initial line of column names. If the lines in such a
    file are terminated by carriage return/newline pairs, the
    statement shown here illustrates the field- and line-handling
    options you would use to load the file:

    LOAD DATA INFILE 'data.txt' INTO TABLE tbl_name
     FIELDS TERMINATED BY ',' ENCLOSED BY '"'
     LINES TERMINATED BY '\r\n'
     IGNORE 1 LINES;

    "If the input values are not necessarily enclosed within
    quotation marks, use OPTIONALLY before the ENCLOSED BY option."

    This implies to me that parsing here is permissive -
    otherwise unambiguous strings without double quotes around
    them will be understood as a string, not rejected.

    Args:
        mysql_load_parameters (MysqlLoadParameters): MySQL load parameters.

    Returns:
        Tuple[Optional[str], Optional[str]]: Tuple of enclosures (
                                            enclosed by, optionally enclosed by).
    """

    mysql_fields_enclosed_by: Optional[str] = None
    mysql_fields_optionally_enclosed_by: Optional[str] = None
    hint_quotechar: HintQuoteChar = mysql_load_parameters.hints.quotechar
    if mysql_load_parameters.hint_quoting == 'all':
        mysql_fields_enclosed_by = hint_quotechar
    elif mysql_load_parameters.hint_quoting == 'minimal':
        mysql_fields_optionally_enclosed_by = hint_quotechar
    elif mysql_load_parameters.hint_quoting == 'nonnumeric':
        mysql_fields_optionally_enclosed_by = hint_quotechar
    elif mysql_load_parameters.hint_quoting is None:
        pass
    else:
        _assert_never(hint_quotechar)
    return (mysql_fields_enclosed_by, mysql_fields_optionally_enclosed_by)


def process_hint_doublequote(mysql_load_parameters: MysqlLoadParameters):
    """
    Process the doublequote hint for the given MySQL load parameters.

    If the field begins with the ENCLOSED BY character, instances of
    that character are recognized as terminating a field value only
    if followed by the field or line TERMINATED BY sequence. To
    avoid ambiguity, occurrences of the ENCLOSED BY character within
    a field value can be doubled and are interpreted as a single
    instance of the character. For example, if ENCLOSED BY '"' is
    specified, quotation marks are handled as shown here:

    We need to ignore flake8's "is vs ==" check because 'is'
    doesn't work currently with MyPy's Literal[] case checking

    Args:
        mysql_load_parameters (MysqlLoadParameters): MySQL load parameters.
    """
    hint_doublequote: HintDoublequote = mysql_load_parameters.hints.doublequote
    if mysql_load_parameters.hint_quoting is not None:

        if hint_doublequote == True:  # noqa: E712
            pass
        elif hint_doublequote == False:  # noqa: E712
            cant_handle_hint(mysql_load_parameters.fail_if_cant_handle_hint,
                             'doublequote',
                             mysql_load_parameters.records_format.hints)
        else:
            _assert_never(hint_doublequote)


def get_escaped_by(mysql_load_parameters: MysqlLoadParameters
                   ) -> Optional[str]:
    """
    Get the escaped by character for the given MySQL load parameters.

    FIELDS ESCAPED BY controls how to read or write special characters:

    * For input, if the FIELDS ESCAPED BY character is not empty,
      occurrences of that character are stripped and the following
      character is taken literally as part of a field value. Some
      two-character sequences that are exceptions, where the first
      character is the escape character.

    [...]

    If the FIELDS ESCAPED BY character is empty, escape-sequence
    interpretation does not occur.

    Args:
        mysql_load_parameters (MysqlLoadParameters): MySQL load parameters.

    Returns:
        Optional[str]: Escaped by character, or None if not present.
    """
    hint_escape: HintEscape = mysql_load_parameters.hints.escape
    if hint_escape is None:
        mysql_fields_escaped_by = None
    elif hint_escape == '\\':
        mysql_fields_escaped_by = '\\'
    else:
        _assert_never(mysql_load_parameters.hint_quoting)
    return mysql_fields_escaped_by


def get_lines_terminated_by(mysql_load_parameters: MysqlLoadParameters) -> str:
    """
    Get the line terminator for the given MySQL load parameters.

    Args:
        mysql_load_parameters (MysqlLoadParameters): MySQL load parameters.

    Returns:
        str: Line terminator.
    """
    hint_record_terminator: HintRecordTerminator = mysql_load_parameters.hints.record_terminator
    mysql_lines_terminated_by = hint_record_terminator
    return mysql_lines_terminated_by


def get_ignore_n_lines(mysql_load_parameters: MysqlLoadParameters) -> int:
    hint_header_row: HintHeaderRow = mysql_load_parameters.hints.header_row
    """
    Get the number of lines to ignore for the given MySQL load parameters.

    We need to ignore flake8's "is vs ==" check because 'is'
    doesn't work currently with MyPy's Literal[] case checking

    Args:
        mysql_load_parameters (MysqlLoadParameters): MySQL load parameters.

    Returns:
        int: Number of lines to ignore.
    """
    if hint_header_row == True:  # noqa: E712
        mysql_ignore_n_lines = 1
    elif hint_header_row == False:  # noqa: E712
        mysql_ignore_n_lines = 0
    else:
        _assert_never(hint_header_row)
    return mysql_ignore_n_lines


def process_hint_compression(mysql_load_parameters: MysqlLoadParameters):
    """
    Process the compression hint for the given MySQL load parameters.

    Args:
        mysql_load_parameters (MysqlLoadParameters): MySQL load parameters.
    """
    hint_compression: HintCompression = mysql_load_parameters.hints.compression
    if hint_compression is not None:
        cant_handle_hint(mysql_load_parameters.fail_if_cant_handle_hint,
                         'compression',
                         mysql_load_parameters.records_format.hints)


def process_temporal_information(mysql_load_parameters: MysqlLoadParameters):
    """
    Process temporal information for the given MySQL load parameters.

    Testing of date/time parsing in MySQL has shown it to be pretty
    conservative.

    That said, for DD/MM and MM/DD support, we may want to look into
    "set trade_date" per
    https://stackoverflow.com/questions/44171283/load-data-local-infile-with-sqlalchemy-and-pymysql

    Args:
        mysql_load_parameters (MysqlLoadParameters): MySQL load parameters.
    """

    if ('YYYY-MM-DD' not in mysql_load_parameters.hints.datetimeformat
            or 'AM' in mysql_load_parameters.hints.datetimeformat):
        cant_handle_hint(mysql_load_parameters.fail_if_cant_handle_hint,
                         'datetimeformat',
                         mysql_load_parameters.records_format.hints)
    if ('YYYY-MM-DD' not in mysql_load_parameters.hints.datetimeformattz
            or 'AM' in mysql_load_parameters.hints.datetimeformattz):
        cant_handle_hint(mysql_load_parameters.fail_if_cant_handle_hint,
                         'datetimeformat',
                         mysql_load_parameters.records_format.hints)
    if mysql_load_parameters.hints.dateformat != 'YYYY-MM-DD':
        cant_handle_hint(mysql_load_parameters.fail_if_cant_handle_hint,
                         'dateformat',
                         mysql_load_parameters.records_format.hints)
    if 'AM' in mysql_load_parameters.hints.timeonlyformat:
        cant_handle_hint(mysql_load_parameters.fail_if_cant_handle_hint,
                         'timeonlyformat',
                         mysql_load_parameters.records_format.hints)


def mysql_load_options(unhandled_hints: Set[str],
                       records_format: DelimitedRecordsFormat,
                       fail_if_cant_handle_hint: bool) -> MySqlLoadOptions:
    """
    The mysql_load_options function takes in unhandled hints, records format,
    and a boolean flag indicating whether to fail if a hint cannot be handled.
    It returns a MySqlLoadOptions object with the options determined from the given parameters.

    Args:

        unhandled_hints (Set[str]): a set of unhandled hints
        records_format (DelimitedRecordsFormat): the format of the records
        fail_if_cant_handle_hint (bool): whether to fail if a hint cannot be handled

    Returns:

        MySqlLoadOptions: a MySqlLoadOptions object with the determined options for loading data
        into MySQL. The object contains the following fields:
            character_set (str): the character set to use
            fields_terminated_by (str): the string used to terminate fields
            fields_enclosed_by (str): the string used to enclose fields
            fields_optionally_enclosed_by (str): the string used to optionally enclose fields
            fields_escaped_by (str): the string used to escape fields
            lines_starting_by (str): the string used to specify lines starting with a specific
                character
            lines_terminated_by (str): the string used to terminate lines
            ignore_n_lines (int): the number of lines to ignore
    """

    mysql_load_parameters = MysqlLoadParameters(fail_if_cant_handle_hint,
                                                records_format,
                                                unhandled_hints)

    mysql_character_set = get_character_set(mysql_load_parameters)
    quiet_remove(mysql_load_parameters.unhandled_hints, 'encoding')

    mysql_fields_terminator = get_field_terminator(mysql_load_parameters)
    quiet_remove(mysql_load_parameters.unhandled_hints, 'field-delimiter')

    mysql_fields_enclosed_by, mysql_fields_optionally_enclosed_by = get_enclosures(
                                                                     mysql_load_parameters)
    quiet_remove(mysql_load_parameters.unhandled_hints, 'quoting')
    quiet_remove(mysql_load_parameters.unhandled_hints, 'quotechar')

    process_hint_doublequote(mysql_load_parameters)
    quiet_remove(mysql_load_parameters.unhandled_hints, 'doublequote')

    mysql_fields_escaped_by = get_escaped_by(mysql_load_parameters)
    quiet_remove(mysql_load_parameters.unhandled_hints, 'escape')

    mysql_lines_starting_by = ''

    mysql_lines_terminated_by = get_lines_terminated_by(mysql_load_parameters)
    quiet_remove(mysql_load_parameters.unhandled_hints, 'record-terminator')

    mysql_ignore_n_lines = get_ignore_n_lines(mysql_load_parameters)
    quiet_remove(mysql_load_parameters.unhandled_hints, 'header-row')

    process_hint_compression(mysql_load_parameters)
    quiet_remove(mysql_load_parameters.unhandled_hints, 'compression')

    process_temporal_information(mysql_load_parameters)
    quiet_remove(mysql_load_parameters.unhandled_hints, 'datetimeformat')
    quiet_remove(mysql_load_parameters.unhandled_hints, 'datetimeformattz')
    quiet_remove(mysql_load_parameters.unhandled_hints, 'dateformat')
    quiet_remove(mysql_load_parameters.unhandled_hints, 'timeonlyformat')

    return MySqlLoadOptions(character_set=mysql_character_set,
                            fields_terminated_by=mysql_fields_terminator,
                            fields_enclosed_by=mysql_fields_enclosed_by,
                            fields_optionally_enclosed_by=mysql_fields_optionally_enclosed_by,
                            fields_escaped_by=mysql_fields_escaped_by,
                            lines_starting_by=mysql_lines_starting_by,
                            lines_terminated_by=mysql_lines_terminated_by,
                            ignore_n_lines=mysql_ignore_n_lines)



from records_mover.db.quoting import quote_schema_and_table, quote_value
from sqlalchemy.engine import Engine


GIG_IN_BYTES = 1024 * 1024 * 1024


def vertica_export_sql(db_engine: Engine,
                       schema: str,
                       s3_url: str,
                       table: str,
                       delimiter: str,
                       record_terminator: str,
                       to_charset: str,
                       # keep these things halfway digestible in memory
                       chunksize: int = 5 * GIG_IN_BYTES) -> str:
    # https://my.vertica.com/docs/8.1.x/HTML/index.htm#Authoring/SQLReferenceManual/Functions/VerticaFunctions/s3export.htm
    template = """
        SELECT S3EXPORT( * USING PARAMETERS {params})
        OVER(PARTITION BEST) FROM {schema_and_table}
    """

    def quote(value: str) -> str:
        return quote_value(db_engine, value)

    params_data = {
        "url": quote(s3_url + 'records.csv'),
        "chunksize": chunksize,
        "to_charset": quote(to_charset),
        "delimiter": quote(delimiter),
        "record_terminator": quote(record_terminator),
    }

    params = ", ".join([f"{key}={value}" for key, value in params_data.items()])
    schema_and_table = quote_schema_and_table(db_engine, schema, table)
    sql = template.format(params=params,
                          schema_and_table=schema_and_table)

    return sql



from records_mover.utils.lazyprop import lazyprop
from records_mover.records.prep import TablePrep, TargetTableDetails
from records_mover.db import DBDriver
from records_mover.pandas import purge_unnamed_unused_columns
from records_mover.records.prep_and_load import prep_and_load
from records_mover.records.schema import RecordsSchema
from records_mover.records.processing_instructions import ProcessingInstructions
from records_mover.records.results import MoveResult
from records_mover.records.targets.table.base import BaseTableMoveAlgorithm
from records_mover.records.sources.dataframes import DataframesRecordsSource
import logging
import sqlalchemy
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .target import TableRecordsTarget  # Dodge circular dependency

logger = logging.getLogger(__name__)


class DoMoveFromDataframesSource(BaseTableMoveAlgorithm):
    def __init__(self,
                 prep: TablePrep,
                 target_table_details: TargetTableDetails,
                 table_target: 'TableRecordsTarget',
                 dfs_source: DataframesRecordsSource,
                 processing_instructions: ProcessingInstructions) -> None:
        self.dfs_source = dfs_source
        self.table_target = table_target

        super().__init__(prep, target_table_details, processing_instructions)

    def move(self) -> MoveResult:
        target_supports_formats = len(self.table_target.known_supported_records_formats()) != 0
        if (target_supports_formats and self.table_target.can_move_from_fileobjs_source()):
            return self.move_from_dataframes_source_via_fileobjs()
        elif (target_supports_formats and
              self.table_target.can_move_from_temp_loc_after_filling_it()):
            # Some databases, like Redshift, can't load from a
            # stream, but can load from files on an object store
            # they're pointed to.
            return self.move_from_dataframes_source_via_temporary_records_directory()
        else:
            logger.info("Known formats for target database: "
                        f"{self.table_target.known_supported_records_formats()}")
            logger.info("Table target can move from fileobjs source? "
                        f"{self.table_target.can_move_from_fileobjs_source()}")
            logger.warning("Loading via INSERT statement as this DB "
                           "driver does not yet support or is not configured for "
                           "more direct load methods.  "
                           "This may be very slow.")
            return self.move_from_dataframes_source_via_insert()

    def move_from_dataframes_source_via_temporary_records_directory(self) -> MoveResult:
        records_format = next(iter(self.table_target.known_supported_records_formats()), None)
        with self.dfs_source.to_fileobjs_source(self.processing_instructions,
                                                records_format) as fileobjs_source:
            return self.table_target.\
                move_from_temp_loc_after_filling_it(fileobjs_source,
                                                    self.processing_instructions)

    def move_from_dataframes_source_via_fileobjs(self) -> MoveResult:
        records_format = next(iter(self.table_target.known_supported_records_formats()), None)
        with self.dfs_source.to_fileobjs_source(self.processing_instructions,
                                                records_format) as fileobjs_source:
            return self.table_target.move_from_fileobjs_source(fileobjs_source,
                                                               self.processing_instructions)

    @lazyprop
    def records_schema(self) -> RecordsSchema:
        return self.dfs_source.initial_records_schema(self.processing_instructions)

    def load(self, driver: DBDriver) -> int:
        rows_loaded = 0
        with self.tbl.db_engine.begin() as db:
            for df in self.dfs_source.dfs:
                df = purge_unnamed_unused_columns(df)
                df = self.records_schema.\
                    assign_dataframe_names(include_index=self.dfs_source.include_index, df=df)
                df.to_sql(name=self.tbl.table_name,
                          con=db,
                          schema=self.tbl.schema_name,
                          index=self.dfs_source.include_index,
                          if_exists='append')
                rows_loaded += len(df.index)
        return rows_loaded

    def move_from_dataframes_source_via_insert(self) -> MoveResult:
        driver = self.tbl.db_driver(self.tbl.db_engine)
        schema_sql = self.records_schema.to_schema_sql(driver,
                                                       self.tbl.schema_name,
                                                       self.tbl.table_name)
        out = prep_and_load(self.tbl, self.prep, schema_sql, self.load,
                            sqlalchemy.exc.InternalError)
        logger.info(f"Loaded {out.move_count} rows into "
                    f"{self.tbl.schema_name}.{self.tbl.table_name} via INSERT statement")
        return out



from enum import Enum
from typing import Dict, Any, Optional, Union, Mapping, List, NoReturn

JsonSchema = Dict[str, Any]

JobConfig = Dict[str, Any]

JsonValue = Optional[Union[bool, str, float, int, Mapping[str, Any], List[Any]]]


# mypy way of validating we're covering all cases of an enum.  This
# version includes an assert at runtime in case poorly-typed input is
# given.
#
# https://github.com/python/mypy/issues/6366#issuecomment-560369716
def _assert_never(x: NoReturn, errmsg: Optional[str] = None) -> NoReturn:
    if errmsg is None:
        errmsg = "Unhandled type: {}".format(type(x).__name__)
    assert False, errmsg


# mypy-friendly way of doing a singleton object:
#
# https://github.com/python/typing/issues/236
class PleaseInfer(Enum):
    token = 1


# mypy way of validating we're covering all cases of an enum.  This
# version allows poorly typed things to pass through at runtime.
#
# https://github.com/python/mypy/issues/6366#issuecomment-560369716
def _ensure_all_cases_covered(x: NoReturn) -> NoReturn:  # type: ignore
    pass



#!/usr/bin/env python

import os
from setuptools import setup, find_packages
from setuptools.command.install import install
from typing import Optional
from decimal import Decimal
from distutils.cmd import Command
import os.path
import sys
import warnings

__version__: Optional[str] = None
# Read in and set version variable without the overhead/requirements
# of the rest of the package.
#
# https://milkr.io/kfei/5-common-patterns-to-version-your-Python-package/5
dir_path = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(dir_path, 'records_mover', 'version.py')) as f:
    exec(f.read())
    assert __version__ is not None


# From https://circleci.com/blog/continuously-deploying-python-packages-to-pypi-with-circleci/
class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""
    description = 'verify that the git tag matches our version'

    def run(self) -> None:
        tag = os.getenv('CIRCLE_TAG')
        tag_formatted_version = f'v{__version__}'

        if tag != tag_formatted_version:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, __version__
            )
            sys.exit(info)


class CoverageRatchetCommand(Command):
    description = 'Run coverage ratchet'
    user_options = []  # type: ignore
    coverage_file: str
    coverage_source_file: str
    coverage_url: str
    type_of_coverage: str

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        """Run command."""
        import xml.etree.ElementTree as ET

        tree = ET.parse(self.coverage_source_file)
        new_coverage = Decimal(tree.getroot().attrib["line-rate"]) * 100

        if not os.path.exists(self.coverage_file):
            with open(self.coverage_file, 'w') as f:
                f.write('0')

        with open(self.coverage_file, 'r') as f:
            high_water_mark = Decimal(f.read())

        if new_coverage < high_water_mark:
            raise Exception(
                f"{self.type_of_coverage} coverage used to be {high_water_mark}; "
                f"down to {new_coverage}%.  Fix by viewing '{self.coverage_url}'")
        elif new_coverage > high_water_mark:
            warnings.warn(
                f"WARNING: Your current coverage is higher than current high watermark.\n"
                f"Consider increasing high watermark in {self.coverage_file} to {new_coverage}.")
        else:
            print(f"Code coverage steady at {new_coverage}%")


class TestCoverageRatchetCommand(CoverageRatchetCommand):
    def initialize_options(self) -> None:
        """Set default values for options."""
        self.type_of_coverage = 'Test'
        self.coverage_url = 'cover/index.html'
        self.coverage_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'metrics',
            'coverage_high_water_mark'
        )
        self.coverage_source_file = "coverage.xml"


class MypyCoverageRatchetCommand(CoverageRatchetCommand):
    def initialize_options(self) -> None:
        """Set default values for options."""
        self.type_of_coverage = 'Mypy'
        self.coverage_url = 'typecover/index.html'
        self.coverage_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            'metrics',
            'mypy_high_water_mark'
        )
        self.coverage_source_file = "typecover/cobertura.xml"


google_api_client_dependencies = [
    # 1.8 seems to be required to use application default creds with
    # 'googleapiclient.discovery.build':
    #
    # https://github.com/googleapis/google-auth-library-python/issues/190
    'google-api-python-client>=1.8.0,<1.9.0',
    #
    # For some reason this issue started happening consistently around 2020-10:
    #
    #  https://app.circleci.com/pipelines/github/bluelabsio/records-mover/1134/workflows/267ca651-def1-4f60-bdfb-0f857a9e0c60/jobs/9770
    #
    # This seems to be a clue:
    #
    #  https://github.com/pypa/pip/issues/8407
    #
    # However, downgrading pip does not seem to resolve it, so it
    # doesn't seem to have been caused by a pip upgrade.  For now
    # we'll explicitly state the dependency for force the install:
    'grpcio<2.0dev,>=1.29.0',
]

pytest_dependencies = [
    'pytest',
    'pytest-cov'
]

itest_dependencies = [
    'jsonschema',  # needed for directory_validator.py
    'pytz',
    'wheel',  # needed to support legacy 'setup.py install'
] + (
    pytest_dependencies +
    # needed for records_database_fixture retrying drop/creates on
    # BigQuery
    google_api_client_dependencies
)

airflow_dependencies = [
    'apache-airflow>=2',
    'apache-airflow-providers-amazon',
    'apache-airflow-providers-google',
]

db_dependencies = [
    'sqlalchemy>=1.4,<2.0',
]

smart_open_dependencies = [
    'smart_open>=2',
]

gcs_dependencies = [
    'google-cloud-storage'
] + smart_open_dependencies

bigquery_dependencies = [
    # This is currently vendored in
    # records_mover/db/postgres/sqlalchemy_postgres_copy.py but
    # once this PR is merged and a new version published, we can
    # use the new upstream version:
    #
    # https://github.com/jmcarp/sqlalchemy-postgres-copy/pull/14
    #
    # 'sqlalchemy-postgres-copy>=0.5,<0.6',
    'sqlalchemy-bigquery',
] + gcs_dependencies + db_dependencies


aws_dependencies = [
    'awscli>=1,<2',
    'boto>=2,<3',
    'boto3',
    's3-concat>=0.1.7,<0.2'
] + smart_open_dependencies

gsheet_dependencies = [
    'google',
    'google_auth_httplib2',
    'oauth2client>=2.0.2,<2.1.0',
    'PyOpenSSL'
] + google_api_client_dependencies

parquet_dependencies = [
    'pyarrow'
]

pandas_dependencies = [
    'pandas>=1.1.5,<2',
]

mysql_dependencies = [
    'pymysql'
] + db_dependencies

redshift_dependencies_base = [
    # sqlalchemy-redshift 0.7.7 introduced support for Parquet in UNLOAD
    'sqlalchemy-redshift>=0.7.7,<0.8.13',
] + aws_dependencies + db_dependencies

redshift_dependencies_binary = [
    'psycopg2-binary',
] + redshift_dependencies_base

redshift_dependencies_source = [
    'psycopg2',
] + redshift_dependencies_base

postgres_depencencies_base = db_dependencies

postgres_dependencies_binary = [
    'psycopg2-binary',
] + postgres_depencencies_base

postgres_dependencies_source = [
    'psycopg2',
] + postgres_depencencies_base

cli_dependencies_base = [
    'odictliteral',
    'jsonschema',
    'docstring_parser',
]

vertica_dependencies = [
    # sqlalchemy-vertica-python 0.5.5 introduced
    # https://github.com/bluelabsio/sqlalchemy-vertica-python/pull/7
    # which fixed a bug pulling schema information from Vertica
    'sqlalchemy-vertica-python>=0.5.5,<0.6',
] + db_dependencies

literally_every_single_database_binary_dependencies = (
    vertica_dependencies +
    postgres_dependencies_binary +
    redshift_dependencies_binary +
    bigquery_dependencies +
    mysql_dependencies
)

typecheck_dependencies = [
    'mypy',
    'lxml',  # needed by mypy HTML coverage reporting
    'sqlalchemy-stubs>=0.3',
    'types-pytz',
    'types-mock',
]

unittest_dependencies = [
    'coverage',
    'mock',
] + (
    pytest_dependencies +
    cli_dependencies_base +
    airflow_dependencies +
    gsheet_dependencies +
    literally_every_single_database_binary_dependencies +
    aws_dependencies +
    pandas_dependencies +
    gcs_dependencies
)

docs_dependencies = [
    'sphinx>=5',  # used to generate and upload docs -
                  # need 5.0 or later for compatibility with other packages
    'sphinx-rtd-theme>=1',  # used to style docs for readthedocs.io
    'sphinx-argparse',  # used to generate documentation of CLI options
    'readthedocs-sphinx-ext>=2',  # also used by readthedocs
    'recommonmark',  # used to be able to use sphinx with markdown
] + (
    # needed for readthedocs.io to be able to evaluate modules with
    # sqlalchemy imports
    db_dependencies +
    # Same with Airflow
    airflow_dependencies +
    # Also boto
    aws_dependencies +
    # Needed to generate docs for CLI options
    cli_dependencies_base
)

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='records-mover',
      version=__version__,  # read right above  # noqa
      description=('Library and CLI to move relational data from one place to another - '
                   'DBs/CSV/gsheets/dataframes/...'),
      long_description=long_description,
      long_description_content_type="text/markdown",
      download_url=f'https://github.com/bluelabsio/records-mover/tarball/{__version__}',  # noqa
      author='Vince Broz',
      author_email='opensource@bluelabs.com',
      packages=find_packages(),
      package_data={
          'records_mover': ['py.typed']
      },
      install_requires=[
          'timeout_decorator',
          # awscli pins PyYAML. We have seen issues at
          # run-time if we don't constrain things here as well. So
          # this tracks the awscli req for python >= 3.6 which is
          # what we support
          #
          # https://github.com/aws/aws-cli/blob/develop/setup.py
          'PyYAML>=3.10,<5.5',
          # Not sure how/if interface will change in db-facts, so
          # let's be conservative about what we're specifying for now.
          'db-facts>=4,<5',
          # Version 4.0.0 is not yet compatible with requests, and pip
          # 20.3 is not smart enough to pay attention to that,
          # resulting in runtime complaints.  Can be removed once 20.4
          # is proven to work.
          'chardet>=3,<4',
          'tenacity>=8.0.1',
          # v5.0.1 resolves https://github.com/exhuma/config_resolver/issues/69
          'config-resolver>=5.0.1,<6',
          'typing_inspect',
          'typing-extensions',
      ],
      extras_require={
          'airflow': airflow_dependencies,
          'db': db_dependencies,
          'gsheets': gsheet_dependencies,
          'cli': cli_dependencies_base,
          'bigquery': bigquery_dependencies,
          'aws': aws_dependencies,
          'mysql': mysql_dependencies,
          'redshift-binary': redshift_dependencies_binary,
          'redshift-source': redshift_dependencies_source,
          'postgres-binary': postgres_dependencies_binary,
          'postgres-source': postgres_dependencies_source,
          'vertica': vertica_dependencies,
          'pandas': pandas_dependencies,
          # don't let it be said we didn't warn you.
          'literally_every_single_database_binary':
          literally_every_single_database_binary_dependencies,
          'itest': itest_dependencies,
          'unittest': unittest_dependencies,
          'typecheck': typecheck_dependencies,
          'gcs': gcs_dependencies,
          'parquet': parquet_dependencies,
          'docs': docs_dependencies,
      },
      entry_points={
          'console_scripts': 'mvrec = records_mover.records.cli:main',
      },
      cmdclass={
          'coverage_ratchet': TestCoverageRatchetCommand,
          'mypy_ratchet': MypyCoverageRatchetCommand,
          'verify': VerifyVersionCommand,
      },
      license='Apache Software License',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Topic :: Database :: Front-Ends',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
      ])



from config_resolver import get_config
from db_facts.db_facts_types import DBFacts
from .creds.base_creds import BaseCreds
from .records.records import Records
from .url.base import BaseFileUrl, BaseDirectoryUrl
from typing import Union, Optional, IO
from .url.resolver import UrlResolver
from records_mover.creds.creds_via_lastpass import CredsViaLastPass
from records_mover.creds.creds_via_airflow import CredsViaAirflow
from records_mover.creds.creds_via_env import CredsViaEnv
from records_mover.logging import set_stream_logging
from records_mover.mover_types import PleaseInfer
import os
import sys
import logging
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .db import DBDriver  # noqa
    from sqlalchemy.engine import Engine, Connection  # noqa
    import boto3  # noqa
    import google.cloud.storage  # noqa


logger = logging.getLogger(__name__)


def _infer_session_type() -> str:
    if 'RECORDS_MOVER_SESSION_TYPE' in os.environ:
        return os.environ['RECORDS_MOVER_SESSION_TYPE']

    config_result = get_config('records_mover', 'bluelabs')
    cfg = config_result.config
    if 'session' in cfg:
        session_cfg = cfg['session']
        session_type: Optional[str] = session_cfg.get('session_type')
        if session_type is not None:
            logger.info(f"Using session_type={session_type} from config file")
            return session_type

    if 'AIRFLOW__CORE__EXECUTOR' in os.environ:
        # Guess based on an env variable sometimes set by Airflow
        return 'airflow'

    return 'env'


def _infer_default_aws_creds_name(session_type: str) -> Optional[str]:
    if session_type == 'airflow':
        return 'aws_default'
    return None


def _infer_default_gcp_creds_name(session_type: str) -> Optional[str]:
    if session_type == 'airflow':
        return 'google_cloud_default'
    return None


def _infer_creds(session_type: str,
                 default_db_creds_name: Optional[str],
                 default_aws_creds_name: Optional[str],
                 default_gcp_creds_name: Optional[str],
                 default_db_facts: Union[PleaseInfer, DBFacts],
                 default_boto3_session: Union[PleaseInfer,
                                              'boto3.session.Session',
                                              None],
                 default_gcp_creds: Union[PleaseInfer,
                                          'google.auth.credentials.Credentials',
                                          None],
                 default_gcs_client: Union[PleaseInfer,
                                           'google.cloud.storage.Client',
                                           None],
                 scratch_s3_url: Union[PleaseInfer,
                                       str,
                                       None],
                 scratch_gcs_url: Union[PleaseInfer,
                                        str,
                                        None]) -> BaseCreds:
    if session_type == 'airflow':
        return CredsViaAirflow(default_db_creds_name=default_db_creds_name,
                               default_aws_creds_name=default_aws_creds_name,
                               default_gcp_creds_name=default_gcp_creds_name,
                               default_db_facts=default_db_facts,
                               default_boto3_session=default_boto3_session,
                               default_gcp_creds=default_gcp_creds,
                               default_gcs_client=default_gcs_client,
                               scratch_s3_url=scratch_s3_url,
                               scratch_gcs_url=scratch_gcs_url)
    elif session_type == 'cli':
        return CredsViaEnv(default_db_creds_name=default_db_creds_name,
                           default_aws_creds_name=default_aws_creds_name,
                           default_gcp_creds_name=default_gcp_creds_name,
                           default_db_facts=default_db_facts,
                           default_boto3_session=default_boto3_session,
                           default_gcp_creds=default_gcp_creds,
                           default_gcs_client=default_gcs_client)
    elif session_type == 'lpass':
        return CredsViaLastPass(default_db_creds_name=default_db_creds_name,
                                default_aws_creds_name=default_aws_creds_name,
                                default_gcp_creds_name=default_gcp_creds_name,
                                default_db_facts=default_db_facts,
                                default_boto3_session=default_boto3_session,
                                default_gcp_creds=default_gcp_creds,
                                default_gcs_client=default_gcs_client,
                                scratch_s3_url=scratch_s3_url,
                                scratch_gcs_url=scratch_gcs_url)
    elif session_type == 'itest':
        return CredsViaEnv(default_db_creds_name=default_db_creds_name,
                           default_aws_creds_name=default_aws_creds_name,
                           default_gcp_creds_name=default_gcp_creds_name,
                           default_db_facts=default_db_facts,
                           default_boto3_session=default_boto3_session,
                           default_gcp_creds=default_gcp_creds,
                           default_gcs_client=default_gcs_client,
                           scratch_s3_url=scratch_s3_url,
                           scratch_gcs_url=scratch_gcs_url)
    elif session_type == 'env':
        return CredsViaEnv(default_db_creds_name=default_db_creds_name,
                           default_aws_creds_name=default_aws_creds_name,
                           default_gcp_creds_name=default_gcp_creds_name,
                           default_db_facts=default_db_facts,
                           default_boto3_session=default_boto3_session,
                           default_gcp_creds=default_gcp_creds,
                           default_gcs_client=default_gcs_client,
                           scratch_s3_url=scratch_s3_url,
                           scratch_gcs_url=scratch_gcs_url)
    elif session_type is not None:
        raise ValueError("Valid session types: cli, lpass, airflow, itest, env - "
                         "consider upgrading records-mover if you're looking for "
                         f"{session_type}.")


class Session():
    def __init__(self,
                 default_db_creds_name: Optional[str] = None,
                 default_aws_creds_name: Union[None, str, PleaseInfer] = PleaseInfer.token,
                 default_gcp_creds_name: Union[None, str, PleaseInfer] = PleaseInfer.token,
                 session_type: Union[str, PleaseInfer] = PleaseInfer.token,
                 scratch_s3_url: Union[None, str, PleaseInfer] = PleaseInfer.token,
                 creds: Union[BaseCreds, PleaseInfer] = PleaseInfer.token,
                 default_db_facts: Union[PleaseInfer, DBFacts] = PleaseInfer.token,
                 default_boto3_session: Union[PleaseInfer,
                                              'boto3.session.Session',
                                              None] = PleaseInfer.token,
                 default_gcp_creds: Union[PleaseInfer,
                                          'google.auth.credentials.Credentials',
                                          None] = PleaseInfer.token,
                 default_gcs_client: Union[PleaseInfer,
                                           'google.cloud.storage.Client',
                                           None] = PleaseInfer.token,
                 scratch_gcs_url: Union[None, str, PleaseInfer] = PleaseInfer.token) -> None:
        """This is an object which ties together configuration on how to do
        key things in order to move records.

        It tries to autoconfigure as much as possible - in many cases
        you won't need to specify any constructor arguments at all.

        Generally unless otherwise configured, this class will look up
        and use the default credentials for things like AWS and GCP if
        they exist and are needed for an operation.  When running in a
        managed environment like Apache Airflow (session_type =
        "airflow"), that might mean looking up an Airflow Connection
        via the Airflow Python API.  On the command line (session_type
        = "cli"), that might mean using e.g., the AWS or GCP Python
        APIs to pull any default credentials which have been
        configured.  In other environments (e.g., containerized
        systems) you may way want to use environment variables
        whenever possible to specify exactly what is desired
        (sesssion_type = 'env').

        :param default_db_creds_name: Name of the database credential to used when
           :meth:`records_mover.Session.get_default_db_engine` is called.  If not specified, the
           default will depend on the session type.
        :param default_aws_creds_name: Name of the AWS IAM credential to used when needed, e.g.
           when reading or writing to an s3:// URL.  This will be inferred unless directly
           specified.
        :param default_gcp_creds_name: Name of the GCP Cloud IAM credential to used when needed,
           e.g. when reading or writing to an gs:// URL.  This will be inferred unless directly
           specified.
        :param session_type: What assumptions to use when inferring and/or looking up credentials.
           Valid values of "airflow" (for code running in Apache Airflow), "cli" (for running on
           the command-line", "lpass" (for using the LastPass password manager for credentials),
           and 'env' (for looking up credentials via environment variables).  This will be inferred
           unless directly specified.
        :param scratch_s3_url: An s3:// URL used as a base directory where temporary
           files/directories can be created.  This is necessary for Amazon Redshift, which supports
           only S3 for bulk import/export.
        :param default_db_facts: Information about the database connection that should be made.
           This is a dictionary with string keys of type `DBFacts
           <https://db-facts.readthedocs.io/en/latest/db_facts.html#db_facts.DBFacts>`_
        :param default_boto3_session: The boto3.Session object used when needed, e.g. when reading
           or writing to an s3:// URL.  This will be inferred unless directly specified.
        :param default_gcp_creds: The google.auth.credentials.Credentials object to used when
           needed, e.g. when reading or writing to an gs:// URL.  This will be inferred unless
           directly specified.
        :param default_gcs_client: The google.cloud.storage.Client object to be used when needed,
           e.g. when reading or writing to an gs:// URL.  This will be inferred unless directly
           specified.
        :param creds: Experimental interface; do not use.
        :param scratch_gcs_url: A gs:// URL used as a base directory where temporary
           files/directories can be created.  This can be helpful for large imports into
           Google BigQuery.
        """
        if session_type is PleaseInfer.token:
            session_type = _infer_session_type()

        if default_aws_creds_name is PleaseInfer.token:
            default_aws_creds_name = _infer_default_aws_creds_name(session_type)

        if default_gcp_creds_name is PleaseInfer.token:
            default_gcp_creds_name = _infer_default_gcp_creds_name(session_type)

        if creds is PleaseInfer.token:
            creds = _infer_creds(session_type,
                                 default_db_creds_name=default_db_creds_name,
                                 default_aws_creds_name=default_aws_creds_name,
                                 default_gcp_creds_name=default_gcp_creds_name,
                                 default_db_facts=default_db_facts,
                                 default_boto3_session=default_boto3_session,
                                 default_gcp_creds=default_gcp_creds,
                                 default_gcs_client=default_gcs_client,
                                 scratch_s3_url=scratch_s3_url,
                                 scratch_gcs_url=scratch_gcs_url)

        self.creds = creds

    @property
    def url_resolver(self) -> UrlResolver:
        return UrlResolver(boto3_session_getter=self.creds.default_boto3_session,
                           gcp_credentials_getter=self.creds.default_gcs_creds,
                           gcs_client_getter=self.creds.default_gcs_client)

    def get_default_db_engine(self) -> 'Engine':
        """Provide the database object corresponding to the default database
        credentials.  The details of how that credential is looked up
        depends on the session_type determined in the constructor, but
        can be overridden using the default_db_creds_name parameter.

        :return: SQLALchemy Engine object

        """
        from .db.connect import engine_from_db_facts
        db_facts = self.creds.default_db_facts()

        return engine_from_db_facts(db_facts)

    def get_db_engine(self,
                      db_creds_name: str,
                      creds_provider: Optional[BaseCreds] = None) -> 'Engine':
        """Provide a database object corresponding to a given credential name.
        The details of how that credential is looked up depends on the
        session_type determined in the constructor.

        :param db_creds_name: Credential name to look up using the configured credentials provider.
        :return: SQLALchemy Engine object
        """
        from .db.connect import engine_from_db_facts

        if creds_provider is None:
            creds_provider = self.creds
        db_facts = creds_provider.db_facts(db_creds_name)
        return engine_from_db_facts(db_facts)

    def db_driver(self, db: Union['Engine', 'Connection']) -> 'DBDriver':
        from .db.factory import db_driver

        kwargs = {}
        scratch_s3_url = self.creds.default_scratch_s3_url()
        if scratch_s3_url is not None:
            try:
                s3_temp_base_loc = self.directory_url(scratch_s3_url)
                kwargs['s3_temp_base_loc'] = s3_temp_base_loc
            except NotImplementedError:
                logger.debug('boto3 not installed', exc_info=True)

        scratch_gcs_url = self.creds.default_scratch_gcs_url()
        if scratch_gcs_url is not None:
            try:
                gcs_temp_base_loc = self.directory_url(scratch_gcs_url)
                kwargs['gcs_temp_base_loc'] = gcs_temp_base_loc
            except NotImplementedError:
                logger.debug('google.cloud.storage not installed', exc_info=True)

        return db_driver(db=db,
                         url_resolver=self.url_resolver,
                         **kwargs)

    def file_url(self, url: str) -> BaseFileUrl:
        return self.url_resolver.file_url(url)

    def directory_url(self, url: str) -> BaseDirectoryUrl:
        return self.url_resolver.directory_url(url)

    def set_stream_logging(self,
                           name: str = 'records_mover',
                           level: int = logging.INFO,
                           stream: IO[str] = sys.stdout,
                           fmt: str = '%(asctime)s - %(message)s',
                           datefmt: str = '%H:%M:%S') -> None:
        """
        records-mover logs details about its operations using Python logging.  This method is a
        simple way to configure that logging to be output to a stream (by default, stdout).

        You can use it for other things (e.g., dependencies of
        records-mover) by adjusting the 'name' argument.

        :param name: Name of the package to set logging under.  If set to 'foo', you can set a log
           variable FOO_LOG_LEVEL to the log level threshold you'd like to set (INFO/WARNING/etc) -
           so you can by default set, say, export RECORDS_MOVER_LOG_LEVEL=WARNING to quiet down
           loging, or export RECORDS_MOVER_LOG_LEVEL=DEBUG to increase it.
        :param level: Logging more detailed than this will not be output to the stream.
        :param stream: Stream which logging should be sent (e.g., sys.stdout, sys.stdin, or perhaps
           a file you open)
        :param fmt: Logging format to send to Python'slogging.Formatter() - determines what details
           will be sent.
        :param datefmt: Date format to send to Python'slogging.Formatter() - determines how the
           current date/time will be recorded in the log.
        """
        set_stream_logging(name=name,
                           level=level,
                           stream=stream,
                           fmt=fmt,
                           datefmt=datefmt)

    @property
    def records(self) -> Records:
        """Property containing a :class:`records_mover.Records` object
        pre-configured with configuration using this Session.  Once you have a
        Session object constructed, this is your jumping off point to moving
        records."""
        return Records(db_driver=self.db_driver,
                       url_resolver=self.url_resolver)



from .base import SupportsToDataframesSource
from contextlib import contextmanager
import itertools
import google.auth.credentials
import collections
from apiclient.discovery import build
import google_auth_httplib2
import httplib2
from typing import Any, Iterator, Optional, Callable, Iterable, List
from .dataframes import DataframesRecordsSource
from ..processing_instructions import ProcessingInstructions

SheetsService = Any


class GoogleSheetsRecordsSource(SupportsToDataframesSource):
    def __init__(self,
                 spreadsheet_id: str,
                 sheet_name_or_range: str,
                 google_cloud_creds: google.auth.credentials.Credentials,
                 out_of_band_column_headers: Optional[Iterable[str]] = None,
                 header_translator: Optional[Callable[[str], str]] = None) -> None:
        self.spreadsheet_id = spreadsheet_id
        self.sheet_name_or_range = sheet_name_or_range
        self.google_cloud_creds = google_cloud_creds
        self.out_of_band_column_headers = out_of_band_column_headers
        self.header_translator = header_translator

    def _authorize(self) -> google_auth_httplib2.AuthorizedHttp:
        credentials = self.google_cloud_creds
        http = httplib2.Http()
        authed_http = google_auth_httplib2.AuthorizedHttp(
            credentials, http=http)
        return authed_http

    def _get_service(self) -> SheetsService:
        http_authorized = self._authorize()
        return build('sheets', 'v4', http=http_authorized)

    @contextmanager
    def to_dataframes_source(self,
                             processing_instructions: ProcessingInstructions) \
            -> Iterator[DataframesRecordsSource]:
        from pandas import DataFrame

        service = self._get_service()

        #
        # https://developers.google.com/sheets/api/reference/rest/v4/ValueRenderOption
        #
        # Values will be calculated, but not formatted in the
        # reply. For example, if A1 is 1.23 and A2 is =A1 and
        # formatted as currency, then A2 would return the number 1.23.
        value_render_option = 'UNFORMATTED_VALUE'
        #
        # https://developers.google.com/sheets/api/reference/rest/v4/DateTimeRenderOption
        #
        # Instructs date, time, datetime, and duration fields to be
        # output as strings in their given number format (which is
        # dependent on the spreadsheet locale).
        date_time_render_option = 'FORMATTED_STRING'

        result = service.spreadsheets().values().get(
            spreadsheetId=self.spreadsheet_id,
            range=self.sheet_name_or_range,
            majorDimension="COLUMNS",  # pull in col order for easy and efficient pandasing
            valueRenderOption=value_render_option,
            dateTimeRenderOption=date_time_render_option,
        ).execute()
        values = result.get('values', [])

        column_headers: Iterable[str]
        data_by_column: List[List[Any]]
        if self.out_of_band_column_headers is not None:
            column_headers = self.out_of_band_column_headers
            data_by_column = values
        else:
            # Use column headers from sheet
            #
            # Apply column header translator if one was passed in

            raw_column_headers = [str(col[0]) for col in values]
            # https://stackoverflow.com/questions/9835762/how-do-i-find-the-duplicates-in-a-list-and-create-another-list-with-them
            duplicates = [
                item
                for item, count in collections.Counter(raw_column_headers).items()
                if count > 1
            ]
            if duplicates:
                raise TypeError("The following column names are duplicated - "
                                f"please make them unique: {duplicates}")
            column_headers = [
                self.header_translator(header)
                for header in raw_column_headers
            ] if self.header_translator else raw_column_headers
            data_by_column = [col[1:] for col in values]

        # Generate data frame from raw google sheet results
        df = DataFrame(
            list(itertools.zip_longest(*[
                map(lambda x: None if x == '' else x, col) for col in data_by_column
            ])),
            columns=column_headers
        )
        yield DataframesRecordsSource(dfs=[df],
                                      processing_instructions=processing_instructions)



from typing_inspect import is_literal_type, get_args
from abc import ABCMeta, abstractmethod
from .types import HintName, UntypedRecordsHints
from typing import TypeVar, Generic, Type, List, TYPE_CHECKING
if TYPE_CHECKING:
    from records_mover.utils.json_schema import JsonSchemaDocument


HintT = TypeVar('HintT')


class Hint(Generic[HintT], metaclass=ABCMeta):
    def __init__(self,
                 hint_name: HintName,
                 default: HintT,
                 description: str) -> None:
        self.default = default
        self.hint_name = hint_name
        self.description = description

    @abstractmethod
    def validate(self,
                 hints: UntypedRecordsHints,
                 fail_if_cant_handle_hint: bool) -> HintT:
        ...

    @abstractmethod
    def json_schema_document(self) -> 'JsonSchemaDocument':
        ...


class StringHint(Hint[str]):
    def json_schema_document(self) -> 'JsonSchemaDocument':
        from records_mover.utils.json_schema import JsonSchemaDocument

        return JsonSchemaDocument('string',
                                  description=self.description)

    def validate(self,
                 hints: UntypedRecordsHints,
                 fail_if_cant_handle_hint: bool) -> str:
        from .utils import cant_handle_hint

        x: object = hints[self.hint_name]
        if isinstance(x, str):
            return x
        else:
            cant_handle_hint(fail_if_cant_handle_hint=fail_if_cant_handle_hint,
                             hint_name=self.hint_name,
                             hints=hints)
            return self.default


LiteralHintT = TypeVar('LiteralHintT')


class LiteralHint(Hint[LiteralHintT]):
    def __init__(self,
                 type_: Type[LiteralHintT],
                 hint_name: HintName,
                 default: LiteralHintT,
                 description: str) -> None:
        assert is_literal_type(type_), f"{hint_name} is not a Literal[]"
        self.type_ = type_
        self.valid_values: List[LiteralHintT] = list(get_args(type_))
        super().__init__(hint_name=hint_name,
                         default=default,
                         description=description)

    def json_schema_document(self) -> 'JsonSchemaDocument':
        from records_mover.utils.json_schema import JsonSchemaDocument

        json_schema_types = {
            bool: 'boolean',
            str: 'string',
            # Even though Python prints the word NoneType in many
            # error messages, NoneType is not an identifier in
            # Python. It’s not in builtins. You can only reach it with
            # type(None).
            #
            # https://realpython.com/null-in-python/
            type(None): 'null',
        }

        types_set = {
            json_schema_types[type(valid_value)]
            for valid_value in self.valid_values
        }

        if all([t == 'boolean' for t in types_set]):
            # We use Literal[True, False] as a type instead of bool as
            # mypy's exhaustive type matching doesn't work with 'bool'.
            #
            # Let's translate that back to boolean here.
            return JsonSchemaDocument('boolean',
                                      description=self.description)

            pass
        else:
            return JsonSchemaDocument(list(types_set),
                                      enum=self.valid_values,
                                      description=self.description)

    def validate(self,
                 hints: UntypedRecordsHints,
                 fail_if_cant_handle_hint: bool) -> LiteralHintT:
        from .utils import cant_handle_hint

        # MyPy doesn't like looking for a generic optional string
        # in a list of specific optional strings.  It's wrong;
        # that's perfectly safe
        x: object = hints[self.hint_name]
        try:
            i = self.valid_values.index(x)  # type: ignore
            return self.valid_values[i]
        except ValueError:
            # well, sort of safe.
            cant_handle_hint(fail_if_cant_handle_hint=fail_if_cant_handle_hint,
                             hint_name=self.hint_name,
                             hints=hints)
            return self.default



from contextlib import contextmanager
from typing import Union, List, IO, Tuple, Optional, Iterator
from ...url import BaseDirectoryUrl
from ...records.delimited import complain_on_unhandled_hints
import pprint
import sqlalchemy
from ...records.load_plan import RecordsLoadPlan
from ...records.records_format import (
    BaseRecordsFormat, DelimitedRecordsFormat, ParquetRecordsFormat, AvroRecordsFormat
)
from ...records.records_directory import RecordsDirectory
from ...records.processing_instructions import ProcessingInstructions
from ...url.resolver import UrlResolver
from google.cloud.bigquery.dbapi.connection import Connection
from google.cloud.bigquery.client import Client
from google.cloud.bigquery.job import LoadJobConfig
from google.cloud.exceptions import NotFound
from .load_job_config_options import load_job_config
import logging
from ..loader import LoaderFromFileobj
from ..errors import NoTemporaryBucketConfiguration

logger = logging.getLogger(__name__)


class BigQueryLoader(LoaderFromFileobj):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine],
                 url_resolver: UrlResolver,
                 gcs_temp_base_loc: Optional[BaseDirectoryUrl])\
            -> None:
        self.db = db
        self.url_resolver = url_resolver
        self.gcs_temp_base_loc = gcs_temp_base_loc

    def best_scheme_to_load_from(self) -> str:
        return 'gs'

    @contextmanager
    def temporary_gcs_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        if self.gcs_temp_base_loc is None:
            raise NoTemporaryBucketConfiguration('Please provide a scratch GCS URL in your config '
                                                 '(e.g., set SCRATCH_GCS_URL to a gs:// URL)')
        else:
            with self.gcs_temp_base_loc.temporary_directory() as temp_loc:
                yield temp_loc

    def _parse_bigquery_schema_name(self, schema: str) -> Tuple[Optional[str], str]:
        # https://github.com/mxmzdlv/pybigquery/blob/master/pybigquery/sqlalchemy_bigquery.py#L320
        dataset = None
        project = None

        schema_split = schema.split('.')
        if len(schema_split) == 1:
            dataset, = schema_split
        elif len(schema_split) == 2:
            project, dataset = schema_split
        else:
            raise ValueError(f"Could not understand schema name {schema}")

        return (project, dataset)

    def _load_job_config(self, load_plan: RecordsLoadPlan) -> LoadJobConfig:
        target_records_format = load_plan.records_format
        processing_instructions = load_plan.processing_instructions
        unhandled_hints = set()
        if isinstance(target_records_format, DelimitedRecordsFormat):
            unhandled_hints = set(target_records_format.hints.keys())
        job_config = load_job_config(unhandled_hints, load_plan)
        if isinstance(target_records_format, DelimitedRecordsFormat):
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, target_records_format.hints)
        return job_config

    def load_from_fileobj(self, schema: str, table: str,
                          load_plan: RecordsLoadPlan, fileobj: IO[bytes]) -> int:
        logger.info("Loading from fileobj into BigQuery")
        # https://googleapis.github.io/google-cloud-python/latest/bigquery/usage/tables.html#creating-a-table
        connection: Connection =\
            self.db.engine.raw_connection().connection
        # https://google-cloud.readthedocs.io/en/latest/bigquery/generated/google.cloud.bigquery.client.Client.html
        client: Client = connection._client
        project_id, dataset_id = self._parse_bigquery_schema_name(schema)
        # https://googleapis.github.io/google-cloud-python/latest/bigquery/generated/google.cloud.bigquery.job.LoadJobConfig.html
        job_config = self._load_job_config(load_plan)

        logger.info(f"Using BigQuery load options: {job_config.to_api_repr()}")
        # https://googleapis.dev/python/bigquery/latest/generated/google.cloud.bigquery.client.Client.html#google.cloud.bigquery.client.Client.load_table_from_file
        job = client.load_table_from_file(fileobj,
                                          f"{schema}.{table}",
                                          job_config=job_config)

        try:
            job.result()  # Waits for table load to complete.
        except Exception:
            logger.error(f"BigQuery load errors:\n\n{pprint.pformat(job.errors)}\n")
            raise

        logger.info(f"Loaded {job.output_rows} rows into {dataset_id}:{table}")
        assert job.output_rows is not None  # should be populated after job result is obtained
        return job.output_rows

    def load(self,
             schema: str,
             table: str,
             load_plan: RecordsLoadPlan,
             directory: RecordsDirectory) -> Optional[int]:

        if directory.scheme != 'gs':
            if self.has_temporary_loadable_directory_loc():
                with self.temporary_gcs_directory_loc() as temp_gcs_loc:
                    gcs_directory = directory.copy_to(temp_gcs_loc)
                    return self.load(schema=schema,
                                     table=table,
                                     load_plan=load_plan,
                                     directory=gcs_directory)
            else:
                return self.load_from_records_directory_via_fileobj(schema=schema,
                                                                    table=table,
                                                                    load_plan=load_plan,
                                                                    directory=directory)

        logger.info("Loading from records directory into BigQuery")
        # https://googleapis.github.io/google-cloud-python/latest/bigquery/usage/tables.html#creating-a-table
        connection: Connection = self.db.engine.raw_connection().connection
        # https://google-cloud.readthedocs.io/en/latest/bigquery/generated/google.cloud.bigquery.client.Client.html
        client: Client = connection._client
        project_id, dataset_id = self._parse_bigquery_schema_name(schema)
        job_config = self._load_job_config(load_plan)

        try:
            table_obj = client.get_table(f"{schema}.{table}")
        except NotFound:
            logger.error("BigQuery table %s.%s not found", schema, table)
            raise

        all_urls = directory.manifest_entry_urls()

        job = client.load_table_from_uri(all_urls,
                                         f"{schema}.{table}",
                                         # Must match the destination dataset location.
                                         location=table_obj.location,
                                         job_config=job_config)
        try:
            job.result()  # Waits for table load to complete.
        except Exception:
            logger.error(f"BigQuery load errors:\n\n{pprint.pformat(job.errors)}\n")
            raise

        logger.info(f"Loaded {job.output_rows} rows into {dataset_id}:{table}")
        assert job.output_rows is not None  # should be populated after job result is obtained
        return job.output_rows

    def temporary_loadable_directory_scheme(self) -> str:
        return 'gs'

    @contextmanager
    def temporary_loadable_directory_loc(self) -> Iterator[BaseDirectoryUrl]:
        with self.temporary_gcs_directory_loc() as temp_loc:
            yield temp_loc

    def has_temporary_loadable_directory_loc(self) -> bool:
        return self.gcs_temp_base_loc is not None

    def can_load_this_format(self, source_records_format: BaseRecordsFormat) -> bool:
        try:
            processing_instructions = ProcessingInstructions()
            load_plan = RecordsLoadPlan(records_format=source_records_format,
                                        processing_instructions=processing_instructions)
            if isinstance(load_plan.records_format, ParquetRecordsFormat):
                return True
            if isinstance(load_plan.records_format, AvroRecordsFormat):
                return True
            if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
                return False
            unhandled_hints = set(load_plan.records_format.hints.keys())
            load_job_config(unhandled_hints, load_plan)
            complain_on_unhandled_hints(processing_instructions.fail_if_dont_understand,
                                        unhandled_hints, load_plan.records_format.hints)
            return True
        except NotImplementedError:
            return False

    def known_supported_records_formats_for_load(self) -> List[BaseRecordsFormat]:
        return [
            DelimitedRecordsFormat(variant='bigquery'),
            ParquetRecordsFormat(),
            AvroRecordsFormat()
        ]



from records_mover.utils import quiet_remove
from records_mover.records.delimited import cant_handle_hint, ValidatedRecordsHints
from typing import Set
from .mode import CopyOptionsMode
from .common import postgres_copy_options_common
from .types import PostgresCopyOptions, CopyOptionsModeType, _assert_never


def postgres_copy_options_csv(unhandled_hints: Set[str],
                              hints: ValidatedRecordsHints,
                              fail_if_cant_handle_hint: bool,
                              mode: CopyOptionsModeType) ->\
        PostgresCopyOptions:
    postgres_options: PostgresCopyOptions = {}
    # FORMAT
    #
    #  Selects the data format to be read or written: text, csv (Comma
    #  Separated Values), or binary. The default is text.
    #
    postgres_options['format'] = 'csv'

    # NULL
    #
    #  Specifies the string that represents a null value. The default
    #  is \N (backslash-N) in text format, and an unquoted empty
    #  string in CSV format. You might prefer an empty string even in
    #  text format for cases where you don't want to distinguish nulls
    #  from empty strings. This option is not allowed when using
    #  binary format.
    #
    #  Note: When using COPY FROM, any data item that matches this
    #  string will be stored as a null value, so you should make sure
    #  that you use the same string as you used with COPY TO.
    #

    # QUOTE
    #
    #  Specifies the quoting character to be used when a data value is
    #  quoted. The default is double-quote. This must be a single
    #  one-byte character. This option is allowed only when using CSV
    #  format.
    #

    postgres_options['quote'] = hints.quotechar
    quiet_remove(unhandled_hints, 'quotechar')

    # ESCAPE
    #
    #  Specifies the character that should appear before a data
    #  character that matches the QUOTE value. The default is the same
    #  as the QUOTE value (so that the quoting character is doubled if
    #  it appears in the data). This must be a single one-byte
    #  character. This option is allowed only when using CSV format.
    #

    if not hints.doublequote:
        cant_handle_hint(fail_if_cant_handle_hint, 'doublequote', hints)
    else:
        quiet_remove(unhandled_hints, 'doublequote')

    if hints.escape is not None:
        cant_handle_hint(fail_if_cant_handle_hint, 'escape', hints)
    else:
        quiet_remove(unhandled_hints, 'escape')

    # FORCE_QUOTE
    #
    #  Forces quoting to be used for all non-NULL values in each
    #  specified column. NULL output is never quoted. If * is
    #  specified, non-NULL values will be quoted in all columns. This
    #  option is allowed only in COPY TO, and only when using CSV
    #  format.
    if mode is CopyOptionsMode.LOADING:
        if hints.quoting != 'minimal':
            cant_handle_hint(fail_if_cant_handle_hint, 'quoting', hints)
        else:
            quiet_remove(unhandled_hints, 'quoting')
    elif mode is CopyOptionsMode.UNLOADING:
        # The values in each record are separated by the DELIMITER
        # character. If the value contains the delimiter character,
        # the QUOTE character, the NULL string, a carriage return, or
        # line feed character, then the whole value is prefixed and
        # suffixed by the QUOTE character, and any occurrence within
        # the value of a QUOTE character or the ESCAPE character is
        # preceded by the escape character. You can also use
        # FORCE_QUOTE to force quotes when outputting non-NULL values
        # in specific columns.

        if hints.quoting == 'minimal':
            pass  # default
        elif hints.quoting == 'all':
            postgres_options['force_quote'] = '*'
        else:
            cant_handle_hint(fail_if_cant_handle_hint, 'quoting', hints)
    else:
        _assert_never(mode)
    quiet_remove(unhandled_hints, 'quoting')

    # As of the 9.2 release (documentation as of 2019-03-12), there's
    # no statement in the docs on what newline formats are accepted in
    # "CSV" mode:
    #
    # https://www.postgresql.org/docs/9.2/sql-copy.html#AEN67247
    #
    # So let's test and find out!
    #
    # This test file is in UNIX newline mode:
    #
    # $ file tests/integration/resources/delimited-bigquery-with-header.csv
    # tests/integration/resources/delimited-bigquery-with-header.csv: ASCII text
    # $
    # It loads fine with:
    # $ mvrec file2table --source.variant bigquery
    # --source.no_compression tests/integration/resources/delimited-bigquery-with-header.csv
    # --target.existing_table drop_and_recreate dockerized-postgres public bigqueryformat
    # $ unix2mac -n tests/integration/resources/delimited-bigquery-with-header.csv
    # tests/integration/resources/delimited-bigquery-with-header-mac.csv
    # $ mvrec file2table --source.variant bigquery --source.no_compression
    # tests/integration/resources/delimited-bigquery-with-header-mac.csv
    # dockerized-postgres public bigqueryformat # loads fine
    # $ unix2dos -n tests/integration/resources/delimited-bigquery-with-header.csv
    # tests/integration/resources/delimited-bigquery-with-header-dos.csv
    # $ mvrec file2table --source.variant bigquery --source.no_compression
    # tests/integration/resources/delimited-bigquery-with-header-dos.csv
    # --target.existing_table drop_and_recreate
    # dockerized-postgres public bigqueryformat # loads fine

    if mode is CopyOptionsMode.LOADING:
        if hints.record_terminator in ("\n", "\r\n", "\r", None):
            quiet_remove(unhandled_hints, 'record-terminator')
        else:
            cant_handle_hint(fail_if_cant_handle_hint, 'records-terminator', hints)
    elif mode is CopyOptionsMode.UNLOADING:
        # No control for this is given - exports appear with unix
        # newlines.
        if hints.record_terminator == "\n":
            quiet_remove(unhandled_hints, 'record-terminator')
        else:
            cant_handle_hint(fail_if_cant_handle_hint, 'records-terminator', hints)
    else:
        _assert_never(mode)

    if hints.compression is not None:
        cant_handle_hint(fail_if_cant_handle_hint, 'compression', hints)
    else:
        quiet_remove(unhandled_hints, 'compression')

    return postgres_copy_options_common(unhandled_hints,
                                        hints,
                                        fail_if_cant_handle_hint,
                                        postgres_options)



from ..driver import DBDriver
import sqlalchemy
from sqlalchemy.sql import text
from records_mover.db.quoting import quote_schema_and_table
from sqlalchemy.schema import Table, Column
import logging
from typing import Optional, Union, Tuple
from ...url.resolver import UrlResolver
from ...url.base import BaseDirectoryUrl
from .loader import VerticaLoader
from ..loader import LoaderFromFileobj, LoaderFromRecordsDirectory
from .unloader import VerticaUnloader
from ..unloader import Unloader
from ...utils.limits import (INT64_MIN, INT64_MAX,
                             FLOAT64_SIGNIFICAND_BITS,
                             num_digits)


logger = logging.getLogger(__name__)


class VerticaDBDriver(DBDriver):
    def __init__(self,
                 db: Union[sqlalchemy.engine.Connection, sqlalchemy.engine.Engine],
                 url_resolver: UrlResolver,
                 s3_temp_base_loc: Optional[BaseDirectoryUrl] = None,
                 **kwargs: object) -> None:
        super().__init__(db)
        self._vertica_loader = VerticaLoader(url_resolver=url_resolver, db=self.db)
        self._vertica_unloader = VerticaUnloader(s3_temp_base_loc=s3_temp_base_loc, db=db)
        self.url_resolver = url_resolver

    def loader(self) -> Optional[LoaderFromRecordsDirectory]:
        return self._vertica_loader

    def loader_from_fileobj(self) -> LoaderFromFileobj:
        return self._vertica_loader

    def unloader(self) -> Optional[Unloader]:
        return self._vertica_unloader

    def has_table(self, schema: str, table: str) -> bool:
        try:
            sql = f"SELECT 1 from {quote_schema_and_table(self.db, schema, table)} limit 0;"
            self.db.execute(sql)
            return True
        except sqlalchemy.exc.ProgrammingError:
            return False

    def schema_sql(self, schema: str, table: str) -> str:
        sql = text("SELECT EXPORT_OBJECTS('', :schema_and_table, false)")
        result = self.db.execute(sql, schema_and_table=f"{schema}.{table}").fetchall()

        if len(result) == 1:
            return result[0].EXPORT_OBJECTS
        else:
            # maybe a permission error?
            return super().schema_sql(schema, table)

    def table(self,
              schema: str,
              table: str) -> Table:
        # sqlalchemy-vertica driver uses system tables that are
        # suuuuper slow in Vertica databases--this is a compromise
        # that pulls column info but not other things
        #
        # https://docs.sqlalchemy.org/en/latest/core/metadata.html#sqlalchemy.schema.Column
        # https://docs.sqlalchemy.org/en/latest/core/reflection.html#
        #    sqlalchemy.engine.reflection.Inspector.get_columns
        columns = [Column(colinfo['name'],
                          type_=colinfo['type'],
                          nullable=colinfo['nullable'],
                          default=colinfo['default'],
                          **colinfo.get('attrs', {}))
                   for colinfo in self.db_engine.dialect.get_columns(self.db_engine,
                                                                     table,
                                                                     schema=schema)]
        t = Table(table, self.meta, schema=schema, *columns)
        return t

    def integer_limits(self,
                       type_: sqlalchemy.types.Integer) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.sql.sqltypes.INTEGER):
            return (INT64_MIN, INT64_MAX)
        return super().integer_limits(type_)

    def fp_constraints(self,
                       type_: sqlalchemy.types.Float) ->\
            Optional[Tuple[int, int]]:
        if isinstance(type_, sqlalchemy.sql.sqltypes.FLOAT):
            return (64, FLOAT64_SIGNIFICAND_BITS)
        return super().fp_constraints(type_)

    def type_for_integer(self,
                         min_value: Optional[int],
                         max_value: Optional[int]) -> sqlalchemy.types.TypeEngine:
        """Find correct integral column type to fit the given min and max integer values"""

        if min_value is not None and max_value is not None:
            if min_value >= INT64_MIN and max_value <= INT64_MAX:
                return sqlalchemy.sql.sqltypes.INTEGER()
            else:
                num_digits_min = num_digits(min_value)
                num_digits_max = num_digits(max_value)
                digit_count = max(num_digits_min, num_digits_max)
                return self.type_for_fixed_point(precision=digit_count,
                                                 scale=0)
        return super().type_for_integer(min_value, max_value)

    def type_for_floating_point(self,
                                fp_total_bits: int,
                                fp_significand_bits: int) -> sqlalchemy.sql.sqltypes.Numeric:
        if fp_significand_bits > FLOAT64_SIGNIFICAND_BITS:
            logger.warning("Falling back to Vertica 64-bit FLOAT type, as Vertica doesn't support "
                           f"fp_significand_bits>=54 (requested: {fp_significand_bits}")
            return sqlalchemy.sql.sqltypes.Float(precision=FLOAT64_SIGNIFICAND_BITS)
        return super().type_for_floating_point(fp_total_bits=fp_total_bits,
                                               fp_significand_bits=fp_significand_bits)



from records_mover.records.load_plan import RecordsLoadPlan
from records_mover.records.delimited import ValidatedRecordsHints
from records_mover.records.records_format import DelimitedRecordsFormat
import logging
from typing import Set, Tuple, Optional
from .date_input_style import determine_input_date_order_style
from .date_output_style import determine_date_output_style
from .csv import postgres_copy_options_csv
from .text import postgres_copy_options_text
from .types import PostgresCopyOptions, DateOrderStyle, DateOutputStyle
from .mode import CopyOptionsMode


logger = logging.getLogger(__name__)

# https://www.postgresql.org/docs/9.2/sql-copy.html


def needs_csv_format(hints: ValidatedRecordsHints) -> bool:
    # This format option is used for importing and exporting the Comma
    # Separated Value (CSV) file format used by many other programs,
    # such as spreadsheets. Instead of the escaping rules used by
    # PostgreSQL's standard text format, it produces and recognizes
    # the common CSV escaping mechanism.

    # The values in each record are separated by the DELIMITER
    # character. If the value contains the delimiter character, the
    # QUOTE character, the NULL string, a carriage return, or line
    # feed character, then the whole value is prefixed and suffixed by
    # the QUOTE character, and any occurrence within the value of a
    # QUOTE character or the ESCAPE character is preceded by the
    # escape character. You can also use FORCE_QUOTE to force quotes
    # when outputting non-NULL values in specific columns.
    if hints.header_row or (hints.quoting is not None):
        return True

    return False


# unloading
def postgres_copy_to_options(unhandled_hints: Set[str],
                             delimited_records_format: DelimitedRecordsFormat,
                             fail_if_cant_handle_hint: bool) ->\
    Tuple[DateOutputStyle,
          Optional[DateOrderStyle],
          PostgresCopyOptions]:
    hints = delimited_records_format.validate(fail_if_cant_handle_hint=fail_if_cant_handle_hint)

    if needs_csv_format(hints):
        copy_options = postgres_copy_options_csv(unhandled_hints,
                                                 hints,
                                                 fail_if_cant_handle_hint,
                                                 CopyOptionsMode.UNLOADING)
    else:
        copy_options = postgres_copy_options_text(unhandled_hints,
                                                  hints,
                                                  fail_if_cant_handle_hint,
                                                  CopyOptionsMode.UNLOADING)

    date_output_style, date_order_style =\
        determine_date_output_style(unhandled_hints,
                                    hints,
                                    fail_if_cant_handle_hint)

    return (date_output_style, date_order_style, copy_options)


# loading
def postgres_copy_from_options(unhandled_hints: Set[str],
                               load_plan: RecordsLoadPlan) ->\
    Tuple[Optional[DateOrderStyle],
          PostgresCopyOptions]:
    fail_if_cant_handle_hint = load_plan.processing_instructions.fail_if_cant_handle_hint
    if not isinstance(load_plan.records_format, DelimitedRecordsFormat):
        raise NotImplementedError("Not currently able to import "
                                  f"{load_plan.records_format.format_type}")
    hints = load_plan.records_format.validate(fail_if_cant_handle_hint=fail_if_cant_handle_hint)

    if needs_csv_format(hints):
        postgres_copy_options = postgres_copy_options_csv(unhandled_hints,
                                                          hints,
                                                          fail_if_cant_handle_hint,
                                                          CopyOptionsMode.LOADING)
    else:
        postgres_copy_options = postgres_copy_options_text(unhandled_hints,
                                                           hints,
                                                           fail_if_cant_handle_hint,
                                                           CopyOptionsMode.LOADING)

    date_order_style: Optional[DateOrderStyle] =\
        determine_input_date_order_style(unhandled_hints,
                                         hints,
                                         fail_if_cant_handle_hint)

    return (date_order_style, postgres_copy_options)



from typing import Any
import io


class IOBaseWrapper(io.IOBase):
    def __init__(self, addinfourl_object: Any) -> None:
        self.obj = addinfourl_object

    def read(self, *args, **kwargs):
        return self.obj.read(*args, **kwargs)
